from dataclasses import dataclass
from typing import Sequence, Pattern

from elastic_tables.model import Cell


@dataclass(frozen=True)
class Column:
    cells: Sequence[Cell]

    def width(self) -> int:
        if self.cells:
            return max(len(cell) for cell in self.cells)
        else:
            return 0

    def matches(self, pattern: Pattern) -> bool:
        return all(pattern.fullmatch(cell.text) for cell in self.cells)



from io import TextIOBase
from typing import Any, IO

from elastic_tables.filter import Filter


class StreamFilter(TextIOBase):
    def __init__(self, stream: IO):
        self.stream = stream
        self.filter = Filter(self._write_output)

    def __getattr__(self, name: str) -> Any:
        return getattr(self.stream, name)

    def _write_output(self, text: str) -> None:
        self.stream.write(text)

    def write(self, data: str) -> int:
        self.filter.input(data)
        return len(data)

    def flush(self) -> None:
        self.filter.flush()
        self.stream.flush()

    def close(self) -> None:
        self.flush()
        self.stream.close()



from pathlib import Path
import sys
from typing import Optional, TextIO

import typer

from elastic_tables.filter import StreamFilter


def do_filter(file: TextIO, align_numeric: bool, align_space: bool) -> None:
    f = StreamFilter(sys.stdout)
    f.filter.align_numeric = align_numeric
    f.filter.align_space = align_space

    # Read line by line so we can use it in a shell pipeline without blocking
    while (string := file.readline()) != "":
        f.write(string)

    f.flush()


def cli(file_name: Optional[Path] = typer.Argument(None), align_numeric: bool = True, align_space: bool = True) -> None:
    sys.stdout.reconfigure(newline='')

    if file_name is None:
        sys.stdin.reconfigure(newline='')
        do_filter(sys.stdin, align_numeric, align_space)
    else:
        with open(file_name, "r", newline='') as file:
            do_filter(file, align_numeric, align_space)


def main():
    typer.run(cli)


if __name__ == "__main__":
    main()



def left(text: str, width: int) -> str:
    return text.ljust(width)


def right(text: str, width: int) -> str:
    return text.rjust(width)


def center(text: str, width: int) -> str:
    return text.center(width)



from __future__ import annotations
from typing import Optional, Callable

from dataclasses import dataclass

AlignmentFunction = Callable[[str, int], str]


@dataclass(frozen=True)
class Cell:
    text: str
    alignment: Optional[AlignmentFunction] = None

    def render(self, width: int, default_alignment: Optional[AlignmentFunction]) -> str:
        if len(self.text) > width:
            raise ValueError(f"Text too long for width {width}: {self.text}")

        alignment = self.alignment or default_alignment or str.ljust
        return alignment(self.text, width)

    def __len__(self):
        return len(self.text)



from __future__ import annotations

from dataclasses import dataclass, replace
from typing import Sequence, Callable

from elastic_tables.model import Line, Cell, AlignmentFunction


@dataclass(frozen=True)
class Row:
    cells: Sequence[Cell]
    line_terminator: str

    @classmethod
    def from_line(cls, line: Line, separator: str = "\t") -> Row:
        cells = [Cell(text) for text in line.content.split(separator)]
        return cls(cells, line.terminator)

    def __len__(self):
        return len(self.cells)

    def render(self, widths: Sequence[int], alignments: Sequence[AlignmentFunction]) -> str:
        cell_width_alignment = zip(self.cells, widths, alignments)
        rendered_cells = (cell.render(width, alignment) for cell, width, alignment in cell_width_alignment)
        return "".join(rendered_cells) + self.line_terminator

    def map_cells(self, function: Callable[[Cell], Cell]) -> Row:
        cells = [function(cell) for cell in self.cells]
        return replace(self, cells=cells)



from typing import Callable, Iterable, Sequence, List

from elastic_tables.model import Line
from elastic_tables.util.text import split_lines


Callback = Callable[[Iterable[Line]], None]


class LineSplitter:
    def __init__(self, callback: Callback = None):
        self._input_buffer = ""
        self._result_buffer: List[Line] = []

        self.callback: Callback = callback or self._buffer_lines

    ##################
    # Public interface
    ##################

    def input(self, string: str) -> None:
        lines, remainder = split_lines(self._input_buffer + string)
        self._input_buffer = remainder
        self.callback(lines)

    def flush(self) -> None:
        if self._input_buffer != "":
            self.callback([Line(self._input_buffer, "")])
            self._input_buffer = ""

    ###################
    # Internal buffer #
    ###################

    def _buffer_lines(self, lines: Sequence[Line]):
        self._result_buffer.extend(lines)

    def lines(self, clear: bool = True) -> Sequence[Line]:
        lines = self._result_buffer
        if clear:
            self._result_buffer = []
        return lines



from itertools import zip_longest
from typing import Iterable


def grouper(group_size: int, iterable: Iterable, fill_value=None) -> Iterable:
    """Adapted from https://docs.python.org/3/library/itertools.html#itertools-recipes"""

    if group_size <= 0:
        raise ValueError(f"group_size must be postive, is {group_size}")

    args = [iter(iterable)] * group_size
    return zip_longest(*args, fillvalue=fill_value)



import re
from typing import Tuple, Iterator

from elastic_tables.model import Line
from elastic_tables.util.iterable import grouper


def split_lines(string: str) -> Tuple[Iterator[Line], str]:
    """Splits a string into lines and returns a tuple of:
      * An iterator that yields Line instances for terminated lines
      * An str that contains the remainder of the string (potentially empty)

    Recognizes \r and \r\n as line terminators.
    """

    # Examples:
    #     ["foo", "\n", "bar"]
    #     ["foo", "\n", "bar", "\n"]
    parts = re.split(r"(\r?\n)", string)

    # An odd number means that the last line is unterminated
    if len(parts) % 2 != 0:
        (*parts, remainder) = parts
    else:
        remainder = ""

    pairs = grouper(2, parts, None)
    lines = (Line(content, terminator) for content, terminator in pairs)

    return lines, remainder



from dataclasses import replace
from typing import Callable

from elastic_tables.parsing import BlockSplitter, LineSplitter
from elastic_tables.model import Block, Table, Cell
from elastic_tables.util import alignment

Callback = Callable[[str], None]


class Filter:
    def __init__(self, callback: Callback = None):
        self.align_numeric = False
        self.align_space = False

        self._block_splitter = BlockSplitter(self._input_block)
        self._line_splitter = LineSplitter(self._block_splitter.input)

        self._callback = callback or self._buffer_result
        self._result_buffer = ""

    @classmethod
    def filter(cls, text: str, align_numeric: bool = None, align_space: bool = None) -> str:
        filter_ = cls()

        if align_numeric is not None:
            filter_.align_numeric = align_numeric
        if align_space is not None:
            filter_.align_space = align_space

        filter_.input(text)
        filter_.flush()

        return filter_.text()

    ##############
    # Processing #
    ##############

    @staticmethod
    def align_cell_space(cell: Cell) -> Cell:
        # TODO clarify blank cell
        # TODO clarify existing alignment
        if cell.text == " " * len(cell):
            return replace(cell, text=" " * max(len(cell) - 2, 0), alignment = None)
        elif cell.text.startswith(" ") and cell.text.endswith(" "):
            return replace(cell, text=cell.text[1:-1], alignment=alignment.center)
        elif cell.text.startswith(" "):
            return replace(cell, text=cell.text[1:], alignment=alignment.right)
        elif cell.text.endswith(" "):
            return replace(cell, text=cell.text[:-1], alignment=alignment.left)
        else:
            return cell

    def _input_block(self, block: Block) -> None:
        table = Table.from_block(block)
        if self.align_numeric:
            table = table.align_numeric()
        if self.align_space:
            table = table.map_cells(self.align_cell_space)
        text = table.render()
        self._callback("".join(text))

    ####################
    # Public interface #
    ####################

    def input(self, text: str) -> None:
        self._line_splitter.input(text)

    def flush(self) -> None:
        self._line_splitter.flush()
        self._block_splitter.flush()

    ###################
    # Internal buffer #
    ###################

    def _buffer_result(self, text: str) -> None:
        self._result_buffer = self._result_buffer + text

    def text(self, clear: bool = True) -> str:
        text = self._result_buffer
        if clear:
            self._result_buffer = []
        return "".join(text)



from dataclasses import dataclass, field, replace
from itertools import zip_longest
import re
from typing import Sequence, Iterator, Callable

from elastic_tables.model import Row, Cell, Block, Column, AlignmentFunction
from elastic_tables.util.alignment import left, right


numeric_pattern = re.compile(r'\s*[+-]?\d+\s*')


@dataclass(frozen=True)
class Table:
    rows: Sequence[Row]
    column_alignment: Sequence[AlignmentFunction] = field(default=None)

    def __post_init__(self):
        if self.column_alignment is None:
            object.__setattr__(self, "column_alignment", [None] * self.column_count())

    @classmethod
    def from_block(cls, block: Block, separator: str = "\t"):
        rows = [Row.from_line(line, separator) for line in block.lines]
        return Table(rows)

    def column_count(self) -> int:
        if self.rows:
            return max(len(row) for row in self.rows)
        else:
            return 0

    def columns(self) -> Iterator[Column]:
        all_cells = zip_longest(*(row.cells for row in self.rows), fillvalue=Cell(""))
        return (Column(list(column_cells)) for column_cells in all_cells)

    def render(self) -> Iterator[str]:
        column_widths = [column.width() for column in self.columns()]
        return (row.render(column_widths, self.column_alignment) for row in self.rows)

    #############
    # Transform #
    #############

    def numeric_alignment(self) -> Sequence[AlignmentFunction]:
        column_is_numeric = [column.matches(numeric_pattern) for column in self.columns()]
        return [right if numeric else left for numeric in column_is_numeric]

    def align_numeric(self) -> "Table":
        column_alignment = self.numeric_alignment()
        return replace(self, column_alignment=column_alignment)

    def map_cells(self, function: Callable[[Cell], Cell]) -> "Table":
        rows = [row.map_cells(function) for row in self.rows]
        return replace(self, rows=rows)



from dataclasses import replace
from typing import Callable, Iterable, Sequence, List

from elastic_tables.model import Line, Block


Callback = Callable[[Block], None]


class BlockSplitter:
    split_on_blank_line = True
    split_on_vertical_tab = True

    def __init__(self, callback: Callable[[Block], None] = None):
        self._input_buffer: List[Line] = []
        self._result_buffer: List[Block] = []

        self._callback: Callback = callback or self._buffer_block

    ##############
    # Processing #
    ##############

    def _add_line(self, line: Line) -> None:
        split_before = False
        split_after = False

        if self.split_on_blank_line:
            if len(line.content.strip()) == 0:
                split_after = True

        if self.split_on_vertical_tab:
            if line.content.startswith("\v"):
                line = replace(line, content=line.content[1:])
                split_before = True

            if line.content.endswith("\v"):
                line = replace(line, content=line.content[:-1])
                split_after = True

        if split_before:
            self.flush()

        self._input_buffer.append(line)

        if split_after:
            self.flush()

    ####################
    # Public interface #
    ####################

    def input(self, lines: Iterable[Line]) -> None:
        for line in lines:
            self._add_line(line)

    def flush(self) -> None:
        if len(self._input_buffer) != 0:
            self._callback(Block(self._input_buffer))
            self._input_buffer = []

    ###################
    # Internal buffer #
    ###################

    def _buffer_block(self, block: Block):
        self._result_buffer.append(block)

    def blocks(self, clear: bool = True) -> Sequence[Block]:
        blocks = self._result_buffer
        if clear:
            self._result_buffer = []
        return blocks
